import os


def current_git_dir():
    """Locate the .git directory."""
    path = os.path.abspath(os.curdir)
    while path != '/':
        if os.path.isdir(os.path.join(path, '.git')):
            return os.path.join(path, '.git')
        path = os.path.dirname(path)
    return None


def list_files(path):
    """Recursively collects a list of files at a path."""
    files = []
    if os.path.isdir(path):
        for stats in os.walk(path):
            for f in stats[2]:
                files.append(os.path.join(stats[0], f))
    elif os.path.isfile(path):
        files = [path]
    return files



import pkg_resources

from therapist.plugins import Plugin
from therapist.plugins.exc import InvalidPlugin, PluginNotInstalled


def list_plugins():
    return [entry_point.name for entry_point in pkg_resources.iter_entry_points(group='therapist.plugin')]


def load_plugin(name):
    for entry_point in pkg_resources.iter_entry_points(group='therapist.plugin', name=name):
        plugin = entry_point.load()

        if issubclass(plugin, Plugin):
            return plugin
        else:
            raise InvalidPlugin('Plugin must inherit from `Plugin`.')
    raise PluginNotInstalled('Plugin `{}` has not been installed.'.format(name))



import time

from xml.etree import ElementTree

from therapist.collection import Collection
from therapist.process import Process


class Result(object):
    SUCCESS = 0
    FAILURE = 1
    ERROR = 2

    def __init__(self, process, status=None):
        self._process = None
        self.process = process
        self.status = status
        self.output = None
        self.error = None
        self.start_time = time.time()
        self.end_time = self.start_time

    def __str__(self):
        if self.is_success:
            return 'SUCCESS'
        elif self.is_failure:
            return 'FAILURE'
        elif self.is_error:
            return 'ERROR'
        else:
            return 'SKIP'

    def __repr__(self):  # pragma: no cover
        return '<Result {}>'.format(self.process)

    @property
    def process(self):
        return self._process

    @process.setter
    def process(self, value):
        if not (isinstance(value, Process)):
            raise TypeError('Expected a `Process` object.')

        self._process = value

    @property
    def is_success(self):
        return self.status == self.SUCCESS

    @property
    def is_failure(self):
        return self.status == self.FAILURE

    @property
    def is_error(self):
        return self.status == self.ERROR

    @property
    def is_skip(self):
        return not (self.is_success or self.is_failure or self.is_error)

    @property
    def execution_time(self):
        return self.end_time - self.start_time

    def mark_complete(self, status=None, output=None, error=None):
        if status is not None:
            self.status = status
        self.output = output
        self.error = error
        self.end_time = time.time()


class ResultCollection(Collection):
    class Meta:
        object_class = Result

    @property
    def has_success(self):
        for result in self.objects:
            if result.is_success:
                return True
        return False

    @property
    def has_failure(self):
        for result in self.objects:
            if result.is_failure:
                return True
        return False

    @property
    def has_skip(self):
        for result in self.objects:
            if result.is_skip:
                return True
        return False

    @property
    def has_error(self):
        for result in self.objects:
            if result.is_error:
                return True
        return False

    @property
    def execution_time(self):
        return sum(result.execution_time for result in self.objects)

    def count(self, **kwargs):
        if 'status' in kwargs:
            return sum(1 for result in self.objects if result.status == kwargs.get('status'))
        return len(self.objects)

    def dump(self):
        """Returns the results in string format."""
        text = ''
        for result in self.objects:
            if result.is_failure or result.is_error:
                text += '\n#{red}#{bright}'
                text += '{}\n'.format(''.ljust(79, '='))

                status = 'FAILED' if result.is_failure else 'ERROR'
                text += '{}: {}\n'.format(status, result.process)
                text += '{}\n#{{reset_all}}'.format(''.ljust(79, '='))

                if result.error:
                    text = '{}{}'.format(text, result.error)
                else:
                    text = '{}{}'.format(text, result.output)

                if not text.endswith('\n'):
                    text += '\n'

        return text

    def dump_junit(self):
        """Returns a string containing XML mapped to the JUnit schema."""
        testsuites = ElementTree.Element('testsuites', name='therapist', time=str(round(self.execution_time, 2)),
                                         tests=str(self.count()), failures=str(self.count(status=Result.FAILURE)),
                                         errors=str(self.count(status=Result.ERROR)))

        for result in self.objects:
            failures = '1' if result.is_failure else '0'
            errors = '1' if result.is_error else '0'

            testsuite = ElementTree.SubElement(testsuites, 'testsuite', id=result.process.name,
                                               name=str(result.process), time=str(round(result.execution_time, 2)),
                                               tests='1', failures=failures, errors=errors)

            testcase = ElementTree.SubElement(testsuite, 'testcase', time=str(round(result.execution_time, 2)))
            testcase.attrib['name'] = result.process.name

            if result.is_failure or result.is_error:
                if result.is_failure:
                    element = ElementTree.SubElement(testcase, 'failure', type='failure')
                else:
                    element = ElementTree.SubElement(testcase, 'error', type='error')

                if result.error:
                    element.text = result.error
                else:
                    element.text = result.output if result.output else ''

        xmlstr = ElementTree.tostring(testsuites, encoding='utf-8').decode('utf-8')
        return '<?xml version="1.0" encoding="UTF-8"?>\n{}'.format(xmlstr)



class Collection(object):
    """A generic iterable set of objects."""
    class Meta:
        pass

    def __init__(self, objects=None):
        self._objects = []

        if objects:
            for obj in objects:
                self.append(obj)

    def __iter__(self):
        return iter(self._objects)

    def __getitem__(self, item):
        return self._objects[item]

    def __str__(self):  # pragma: no cover
        return str(self._objects)

    def __repr__(self):  # pragma: no cover
        return '<{}>'.format(self.__class__.__name__)

    def __bool__(self):
        return bool(self._objects)
    __nonzero__ = __bool__

    @property
    def objects(self):
        return self._objects

    def append(self, v):
        object_class = getattr(self.Meta, 'object_class', object)

        if not isinstance(v, object_class):
            raise TypeError('Expected an instance of `{}`.'.format(object_class.__name__))

        self._objects.append(v)



import subprocess

from therapist.collection import Collection
from therapist.exc import Error
from therapist.process import Process
from therapist.runner.result import Result


class Action(Process):
    def execute(self, **kwargs):
        files = self.filter_files(kwargs.get('files'))

        result = Result(self)

        if 'run' in self.config and files:
            run_command = self.config.get('run').format(files=' '.join(files))

            try:
                pipes = subprocess.Popen(run_command, shell=True, cwd=kwargs.get('cwd'), stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE)
            except OSError as err:
                result.mark_complete(status=Result.ERROR, error='OSError {}'.format(str(err)))
            else:
                std_out, std_err = pipes.communicate()

                status = Result.SUCCESS if pipes.returncode == 0 else Result.FAILURE
                result.mark_complete(status=status, output=std_out.decode('utf-8'), error=std_err.decode('utf-8'))

        return result


class ActionCollection(Collection):
    class Meta:
        object_class = Action

    class DoesNotExist(Error):
        pass

    def get(self, name):
        for action in self.objects:
            if action.name == name:
                return action
        raise self.DoesNotExist('`{}` is not a valid action.'.format(name))



import subprocess


def to_cli_args(*args, **kwargs):
    cmd = []
    for k, v in kwargs.items():
        k = k.replace('_', '-')
        short = len(k) == 1
        if short:
            cmd.append('-' + k)
            if v is not True:
                cmd.append(v)
        else:
            if v is True:
                cmd.append('--' + k)
            else:
                cmd.append('--{0}={1}'.format(k, v))

    cmd.extend(args)
    return cmd


class Git(object):
    def __init__(self, cmd=None, repo_path=None):
        self.repo_path = repo_path
        self.current = list(cmd) if cmd else ['git']

    def __call__(self, *args, **kwargs):
        cmd = self.current + to_cli_args(*args, **kwargs)

        subprocess_kwargs = {
            'stdout': subprocess.PIPE,
            'stderr': subprocess.PIPE
        }

        if self.repo_path:
            subprocess_kwargs['cwd'] = self.repo_path

        pipes = subprocess.Popen(cmd, **subprocess_kwargs)
        out, err = pipes.communicate()
        return out.decode('utf-8'), err.decode('utf-8')

    def __getattr__(self, name):
        name = name.replace('_', '-')
        return Git(self.current + [name], repo_path=self.repo_path)

    def __str__(self):  # pragma: no cover
        return str(self.current)

    def __repr__(self):  # pragma: no cover
        return '<Git {}>'.format(str(self))



import os
import yaml

from therapist.exc import Error
from therapist.plugins.exc import InvalidPlugin, PluginNotInstalled
from therapist.plugins.loader import load_plugin
from therapist.plugins.plugin import PluginCollection
from therapist.runner.action import Action, ActionCollection
from therapist.utils.git import Git, Status


class Runner(object):
    class Misconfigured(Error):
        NO_CONFIG_FILE = 1
        NO_ACTIONS_OR_PLUGINS = 2
        EMPTY_CONFIG = 3
        ACTIONS_WRONGLY_CONFIGURED = 4
        PLUGINS_WRONGLY_CONFIGURED = 5
        PLUGIN_NOT_INSTALLED = 6
        PLUGIN_INVALID = 7

        def __init__(self, *args, **kwargs):
            self.code = kwargs.pop('code', None)
            super(self.__class__, self).__init__(*args, **kwargs)

    def __init__(self, cwd, files=None, include_unstaged=False, include_untracked=False,
                 include_unstaged_changes=False):
        self.actions = ActionCollection()
        self.plugins = PluginCollection()
        self.unstaged_changes = False

        self.cwd = os.path.abspath(cwd)
        self.git = Git(repo_path=self.cwd)
        self.include_unstaged_changes = include_unstaged_changes or include_unstaged

        # Try and load the config file
        try:
            with open(os.path.join(self.cwd, '.therapist.yml'), 'r') as f:
                config = yaml.safe_load(f)
        except IOError:
            raise self.Misconfigured('Missing configuration file.', code=self.Misconfigured.NO_CONFIG_FILE)
        else:
            if config is None:
                raise self.Misconfigured('Empty configuration file.', code=self.Misconfigured.EMPTY_CONFIG)

            if 'actions' in config:
                try:
                    actions = config['actions']
                except TypeError:
                    raise self.Misconfigured('`actions` was not configured correctly.',
                                             code=self.Misconfigured.ACTIONS_WRONGLY_CONFIGURED)
                else:
                    if not isinstance(actions, dict):
                        raise self.Misconfigured('`actions` was not configured correctly.',
                                                 code=self.Misconfigured.ACTIONS_WRONGLY_CONFIGURED)

                    for action_name in actions:
                        settings = actions[action_name]
                        if settings is None:
                            settings = {}
                        self.actions.append(Action(action_name, **settings))

            if 'plugins' in config:
                try:
                    plugins = config['plugins']
                except TypeError:
                    raise self.Misconfigured('`plugins` was not configured correctly.',
                                             code=self.Misconfigured.PLUGINS_WRONGLY_CONFIGURED)
                else:
                    if not isinstance(plugins, dict):
                        raise self.Misconfigured('`plugins` was not configured correctly.',
                                                 code=self.Misconfigured.PLUGINS_WRONGLY_CONFIGURED)

                    for plugin_name in plugins:
                        try:
                            plugin = load_plugin(plugin_name)
                        except PluginNotInstalled:
                            raise self.Misconfigured('Plugin `{}` is not installed.'.format(plugin_name),
                                                     code=self.Misconfigured.PLUGIN_NOT_INSTALLED)
                        except InvalidPlugin:
                            raise self.Misconfigured('Plugin `{}` is not a valid plugin.'.format(plugin_name),
                                                     code=self.Misconfigured.PLUGIN_INVALID)
                        settings = plugins[plugin_name]
                        if settings is None:
                            settings = {}
                        self.plugins.append(plugin(plugin_name, **settings))

            if not (self.actions or self.plugins):
                raise self.Misconfigured('`actions` or `plugins` must be specified in the configuration file.',
                                         code=self.Misconfigured.NO_ACTIONS_OR_PLUGINS)

        if files is None:
            files = []

            untracked_files = 'all' if include_untracked else 'no'
            out, err = self.git.status(porcelain=True, untracked_files=untracked_files)

            for line in out.splitlines():
                file_status = Status(line)

                # Check if staged files were modified since being staged
                if file_status.is_staged and file_status.is_modified:
                    self.unstaged_changes = True

                # Skip unstaged files if the `unstaged` flag is False
                if not file_status.is_staged and not include_unstaged and not include_untracked:
                    continue

                # Skip deleted files
                if file_status.is_deleted:
                    continue

                files.append(file_status.path)

        self.files = files

    def run_process(self, process):
        """Runs a single action."""
        message = u'#{bright}'
        message += u'{} '.format(str(process)[:68]).ljust(69, '.')

        if not self.include_unstaged_changes:
            self.git.stash(keep_index=True, quiet=True)

        try:
            result = process(files=self.files, cwd=self.cwd)
        except:
            raise
        finally:
            if not self.include_unstaged_changes:
                self.git.reset(hard=True, quiet=True)
                self.git.stash.pop(index=True, quiet=True)

        if result.is_success:
            message += u' #{green}[SUCCESS]'
        elif result.is_failure:
            message += u' #{red}[FAILURE]'
        elif result.is_skip:
            message += u' #{cyan}[SKIPPED]'
        elif result.is_error:
            message += u' #{red}[ERROR!!]'

        return result, message



import re


class Status(object):
    def __init__(self, status):
        self.state = status[0]
        self.is_modified = status[1] == 'M'

        if self.is_renamed:
            matches = re.search('(\S+?)\s+->\s+(\S+?)$', status[3:])
            self.original_path = matches.groups()[0]
            self.path = matches.groups()[1]
        else:
            self.path = status[3:]

    def __str__(self):
        status = '{:1}{:1}'.format(self.state, 'M' if self.is_modified else '')
        if self.is_renamed:
            status = '{:3}{} -> {}'.format(status, self.original_path, self.path)
        else:
            status = '{:3}{}'.format(status, self.path)
        return status

    def __repr__(self):  # pragma: no cover
        return '<Status {}>'.format(self.path)

    @property
    def is_staged(self):
        """If the state is empty then the file is unstaged."""
        return self.state not in (' ', '?',)

    @property
    def is_untracked(self):
        """Returns true if the file is untracked."""
        return self.state == '?'

    @property
    def is_renamed(self):
        """Returns true if the file is renamed."""
        return self.state == 'R'

    @property
    def is_deleted(self):
        """Returns true if the file is deleted."""
        return self.state == 'D'

    @property
    def is_added(self):
        """Returns true if the file is added."""
        return self.state == 'A'



import hashlib
import os
import re
import shutil
import subprocess

import click
import colorama

from six import print_

from therapist import __version__
from therapist.messages import (NOT_GIT_REPO_MSG, HOOK_ALREADY_INSTALLED_MSG, EXISTING_HOOK_MSG,
                                CONFIRM_PRESERVE_LEGACY_HOOK_MSG, COPYING_HOOK_MSG, DONE_COPYING_HOOK_MSG,
                                CONFIRM_REPLACE_HOOK_MSG, INSTALL_ABORTED_MSG, INSTALLING_HOOK_MSG,
                                DONE_INSTALLING_HOOK_MSG, NO_HOOK_INSTALLED_MSG, UNINSTALL_ABORTED_MSG,
                                CONFIRM_UNINSTALL_HOOK_MSG, CURRENT_HOOK_NOT_THERAPIST_MSG, LEGACY_HOOK_EXISTS_MSG,
                                CONFIRM_RESTORE_LEGACY_HOOK_MSG, COPYING_LEGACY_HOOK_MSG, DONE_COPYING_LEGACY_HOOK_MSG,
                                REMOVING_LEGACY_HOOK_MSG, DONE_REMOVING_LEGACY_HOOK_MSG, UNINSTALLING_HOOK_MSG,
                                DONE_UNINSTALLING_HOOK_MSG, MISCONFIGURED_MSG, UNSTAGED_CHANGES_MSG)
from therapist.plugins.loader import list_plugins
from therapist.runner import Runner
from therapist.runner.result import ResultCollection
from therapist.utils.filesystem import current_git_dir, list_files
from therapist.utils.hook import identify_hook
from therapist.utils.git import Git


BASE_DIR = os.path.abspath(os.path.dirname(__file__))

git = Git()


def output(message, **kwargs):
    def repl(match):  # pragma: no cover
        attr = match.group(0)[2:-1].upper()
        if hasattr(colorama.Fore, attr):
            return getattr(colorama.Fore, attr)
        elif hasattr(colorama.Style, attr):
            return getattr(colorama.Style, attr)
        else:
            return match.group(0)

    message, count = re.subn('#{(.+?)}', repl, message)
    message = colorama.Style.RESET_ALL + message + colorama.Style.RESET_ALL
    print_(message, **kwargs)


@click.group(invoke_without_command=True)
@click.option('--version', '-V', is_flag=True, help='Show the version and exit.')
def cli(version):
    """A smart pre-commit hook for git."""
    if version:
        output('v{}'.format(__version__))


@cli.command()
@click.option('--force', '-f', is_flag=True, help='Force installation of the hook. This will replace any existing hook '
                                                  'unless you also use the --preserve-legacy option.')
@click.option('--no-color', is_flag=True, help='Disables colors and other rich output.')
@click.option('--preserve-legacy', is_flag=True, help='Preserves any existing pre-commit hook.')
def install(**kwargs):
    """Install the pre-commit hook."""
    force = kwargs.get('force')
    preserve_legacy = kwargs.get('preserve_legacy')

    colorama.init(strip=kwargs.get('no_color'))

    stdout = subprocess.check_output('which therapist', shell=True)
    therapist_bin = stdout.decode('utf-8').split()[0]

    git_dir = current_git_dir()

    if git_dir is None:
        output(NOT_GIT_REPO_MSG)
        exit(1)

    with open(os.path.join(BASE_DIR, 'hooks', 'pre-commit-template'), 'r') as f:
        srchook = f.read()
        srchook_hash = hashlib.md5(srchook.encode()).hexdigest()

    dsthook_path = os.path.join(git_dir, 'hooks', 'pre-commit')

    if os.path.isfile(dsthook_path):
        dsthook_hash = identify_hook(dsthook_path)
        if dsthook_hash:
            if dsthook_hash == srchook_hash:
                output(HOOK_ALREADY_INSTALLED_MSG)
                exit(0)
        else:
            if not force and not preserve_legacy:
                print(EXISTING_HOOK_MSG)
                preserve_legacy = click.confirm(CONFIRM_PRESERVE_LEGACY_HOOK_MSG, default=True)

            if preserve_legacy:
                output(COPYING_HOOK_MSG, end='')
                shutil.copy2(dsthook_path, '{}.legacy'.format(dsthook_path))
                output(DONE_COPYING_HOOK_MSG)
            elif not force:
                if not click.confirm(CONFIRM_REPLACE_HOOK_MSG, default=False):
                    output(INSTALL_ABORTED_MSG)
                    exit(1)

    output(INSTALLING_HOOK_MSG, end='')
    with open(dsthook_path, 'w+') as f:
        srchook = srchook.replace('%hash%', srchook_hash)
        srchook = srchook.replace('%therapist_bin%', therapist_bin)
        f.write(srchook)
    os.chmod(dsthook_path, 0o775)
    output(DONE_INSTALLING_HOOK_MSG)


@cli.command()
@click.option('--force', '-f', is_flag=True, help='Force uninstallation of the Therapist pre-commit hook. This will '
                                                  'also remove any legacy hook unless you also use the '
                                                  '--restore-legacy option.')
@click.option('--no-color', is_flag=True, help='Disables colors and other rich output.')
@click.option('--restore-legacy', is_flag=True, help='Restores any legacy pre-commit hook.')
def uninstall(**kwargs):
    """Uninstall the current pre-commit hook."""
    force = kwargs.get('force')
    restore_legacy = kwargs.get('restore_legacy')

    colorama.init(strip=kwargs.get('no_color'))

    git_dir = current_git_dir()

    if git_dir is None:
        output(NOT_GIT_REPO_MSG)
        exit(1)

    hook_path = os.path.join(git_dir, 'hooks', 'pre-commit')

    if not os.path.isfile(hook_path):
        output(NO_HOOK_INSTALLED_MSG)
        exit(0)

    hook_hash = identify_hook(hook_path)

    if hook_hash:
        if not force:
            if not click.confirm(CONFIRM_UNINSTALL_HOOK_MSG, default=False):
                output(UNINSTALL_ABORTED_MSG)
                exit(1)
    else:
        output(CURRENT_HOOK_NOT_THERAPIST_MSG)
        exit(1)

    legacy_hook_path = os.path.join(git_dir, 'hooks', 'pre-commit.legacy')

    if os.path.isfile(legacy_hook_path):
        if not force and not restore_legacy:
            output(LEGACY_HOOK_EXISTS_MSG)
            restore_legacy = click.confirm(CONFIRM_RESTORE_LEGACY_HOOK_MSG, default=True)
        if restore_legacy:
            output(COPYING_LEGACY_HOOK_MSG, end='')
            shutil.copy2(legacy_hook_path, hook_path)
            os.remove(legacy_hook_path)
            output(DONE_COPYING_LEGACY_HOOK_MSG)
            exit(0)
        else:
            if force or click.confirm('Would you like to remove the legacy hook?', default=False):
                output(REMOVING_LEGACY_HOOK_MSG, end='')
                os.remove(legacy_hook_path)
                output(DONE_REMOVING_LEGACY_HOOK_MSG)

    output(UNINSTALLING_HOOK_MSG, end='')
    os.remove(hook_path)
    output(DONE_UNINSTALLING_HOOK_MSG)


@cli.command()
@click.argument('paths', nargs=-1)
@click.option('--action', '-a', default=None, help='A name of a specific action to be run.')
@click.option('--include-unstaged', is_flag=True, help='Include unstaged files.')
@click.option('--include-unstaged-changes', is_flag=True, help='Include unstaged changes to staged files.')
@click.option('--include-untracked', is_flag=True, help='Include untracked files.')
@click.option('--junit-xml', default=None, help='Create a junit-xml style report file at the given path.')
@click.option('--no-color', is_flag=True, help='Disables colors and other rich output.')
@click.option('--plugin', '-p', default=None, help='A name of a specific plugin to be run.')
@click.option('--use-tracked-files', is_flag=True, help='Runs actions against all tracked files.')
@click.option('--quiet', '-q', is_flag=True, help='Suppress all output, unless an error occurs.')
def run(**kwargs):
    """Run the Therapist suite."""
    paths = kwargs.pop('paths', ())
    action = kwargs.pop('action')
    plugin = kwargs.pop('plugin')
    junit_xml = kwargs.pop('junit_xml')
    use_tracked_files = kwargs.pop('use_tracked_files')
    quiet = kwargs.pop('quiet')

    colorama.init(strip=kwargs.pop('no_color'))

    git_dir = current_git_dir()

    if git_dir is None:
        output(NOT_GIT_REPO_MSG)
        exit(1)

    repo_root = os.path.dirname(git_dir)

    files = []
    if paths:
        # We want to look at files in their current state if paths are passed through
        kwargs['include_unstaged_changes'] = True

        # If paths were provided get all the files for each path
        for path in paths:
            for f in list_files(path):
                f = os.path.relpath(f, repo_root)
                if not f.startswith('..'):  # Don't include files outside the repo root.
                    files.append(f)
    elif use_tracked_files:
        # If the use tracked files flag was passed, get a list of all the tracked files
        out, err = git.ls_files()
        files = out.splitlines()

    if files or paths:
        kwargs['files'] = files

    try:
        runner = Runner(repo_root, **kwargs)
    except Runner.Misconfigured as err:
        output(MISCONFIGURED_MSG.format(err.message))

        if err.code == Runner.Misconfigured.PLUGIN_NOT_INSTALLED:
            output('Installed plugins:')
            for p in list_plugins():
                output(p)

        exit(1)
    else:
        results = ResultCollection()

        if runner.unstaged_changes and not quiet:
            output(UNSTAGED_CHANGES_MSG, end='\n\n')

        processes = list(runner.actions) + list(runner.plugins)

        if plugin:
            try:
                processes = [runner.plugins.get(plugin)]
            except runner.plugins.DoesNotExist as err:
                output('{}\nAvailable plugins:'.format(err.message))

                for p in runner.plugins:
                    output(p.name)
                exit(1)

        if action:
            try:
                processes = [runner.actions.get(action)]
            except runner.actions.DoesNotExist as err:
                output('{}\nAvailable actions:'.format(err.message))

                for a in runner.actions:
                    output(a.name)
                exit(1)

        for process in processes:
            result, message = runner.run_process(process)
            results.append(result)

            if not quiet:
                output(message)

        if junit_xml:
            with open(junit_xml, 'w+') as f:
                f.write('{}'.format(results.dump_junit()))

        if not quiet:
            output(results.dump())
            output('#{{bright}}{}\nCompleted in: {}s'.format(''.ljust(79, '-'), round(results.execution_time, 2)))

        if results.has_error:
            exit(1)
        elif results.has_failure:
            exit(2)



from pathspec import PathSpec
from pathspec.patterns import GitWildMatchPattern


class Process(object):
    def __init__(self, name, **kwargs):
        self.name = name
        self.description = kwargs.pop('description', None)

        self._include = None
        self.include = kwargs.pop('include', None)

        self._exclude = None
        self.exclude = kwargs.pop('exclude', None)

        self.config = kwargs

    def __call__(self, **kwargs):
        return self.execute(**kwargs)

    def __str__(self):
        return self.description if self.description else self.name

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.name)

    @property
    def include(self):
        return self._include

    @include.setter
    def include(self, value):
        if value is None:
            self._include = []
        else:
            self._include = value if isinstance(value, list) else [value]

    @property
    def exclude(self):
        return self._exclude

    @exclude.setter
    def exclude(self, value):
        if value is None:
            self._exclude = []
        else:
            self._exclude = value if isinstance(value, list) else [value]

    def filter_files(self, files):
        if self.include:
            spec = PathSpec(map(GitWildMatchPattern, self.include))
            files = list(spec.match_files(files))

        if self.exclude:
            spec = PathSpec(map(GitWildMatchPattern, self.exclude))
            exclude = list(spec.match_files(files))
            files = list(filter(lambda f: f not in exclude, files))

        return files

    def execute(self, **kwargs):
        raise NotImplementedError()  # pragma: no cover



from therapist.collection import Collection
from therapist.exc import Error
from therapist.process import Process


class Plugin(Process):  # pragma: no cover
    def __repr__(self):
        return '<Plugin {}>'.format(self.name)

    def execute(self, **kwargs):
        raise NotImplementedError()


class PluginCollection(Collection):
    class Meta:
        object_class = Plugin

    class DoesNotExist(Error):
        pass

    def get(self, name):
        for plugin in self.objects:
            if plugin.name == name:
                return plugin
        raise self.DoesNotExist('`{}` is not a valid plugin.'.format(name))



NOT_GIT_REPO_MSG = '#{red}Not a git repository (or any of the parent directories)'

HOOK_ALREADY_INSTALLED_MSG = 'The pre-commit hook has already been installed.'

EXISTING_HOOK_MSG = ('#{yellow}There is an existing pre-commit hook.\n'
                     '#{reset_all}Therapist can preserve this legacy hook and run it before the Therapist pre-commit '
                     'hook.')

CONFIRM_PRESERVE_LEGACY_HOOK_MSG = 'Would you like to preserve this legacy hook?'

COPYING_HOOK_MSG = 'Copying `pre-commit` to `pre-commit.legacy`...\t'

DONE_COPYING_HOOK_MSG = '#{green}#{bright}DONE'

CONFIRM_REPLACE_HOOK_MSG = 'Do you want to replace this hook?'

INSTALL_ABORTED_MSG = 'Installation aborted.'

INSTALLING_HOOK_MSG = 'Installing pre-commit hook...\t'

DONE_INSTALLING_HOOK_MSG = '#{green}#{bright}DONE'

NO_HOOK_INSTALLED_MSG = 'There is no pre-commit hook currently installed.'

UNINSTALL_ABORTED_MSG = 'Uninstallation aborted.'

CONFIRM_UNINSTALL_HOOK_MSG = 'Are you sure you want to uninstall the current pre-commit hook?'

CURRENT_HOOK_NOT_THERAPIST_MSG = ('#{yellow}The current pre-commit hook is not the Therapist pre-commit hook.\n'
                                  '#{reset_all}Uninstallation aborted.')

LEGACY_HOOK_EXISTS_MSG = '#{yellow}There is a legacy pre-commit hook present.'

CONFIRM_RESTORE_LEGACY_HOOK_MSG = 'Would you like to restore the legacy hook?'

COPYING_LEGACY_HOOK_MSG = 'Copying `pre-commit.legacy` to `pre-commit`...\t'

DONE_COPYING_LEGACY_HOOK_MSG = '#{green}#{bright}DONE'

REMOVING_LEGACY_HOOK_MSG = 'Removing `pre-commit.legacy`...\t'

DONE_REMOVING_LEGACY_HOOK_MSG = '#{green}#{bright}DONE'

UNINSTALLING_HOOK_MSG = 'Uninstalling pre-commit hook...\t'

DONE_UNINSTALLING_HOOK_MSG = '#{green}#{bright}DONE'

MISCONFIGURED_MSG = '#{{red}}Misconfigured: {}'

UNSTAGED_CHANGES_MSG = '#{yellow}You have unstaged changes.'



"""
A smart pre-commit hook for git.
"""
import os

from setuptools import find_packages, setup


DEPENDENCIES = [
    'click >= 6.7',
    'colorama >= 0.3.7',
    'pathspec >= 0.5.0',
    'PyYAML >= 3.12',
    'six >= 1.10.0',
]

ROOT = os.path.abspath(os.path.dirname(__file__))

version = __import__('therapist').__version__


setup(
    name='therapist',
    version=version,
    url='https://github.com/rehandalal/therapist',
    license='Mozilla Public License Version 2.0',
    author='Rehan Dalal',
    author_email='rehan@meet-rehan.com',
    description='A smart pre-commit hook for git.',
    long_description=__doc__,
    packages=find_packages(exclude=['tests', 'docs']),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=DEPENDENCIES,
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    py_modules=['therapist'],
    entry_points={
        'console_scripts': [
            'therapist = therapist.cli:cli',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing',
    ]
)



def identify_hook(path):
    """Verify that the file at path is the therapist hook and return the hash"""
    with open(path, 'r') as f:
        f.readline()  # Discard the shebang line
        version_line = f.readline()
        if version_line.startswith('# THERAPIST'):
            return version_line.split()[2]
